from smartcard.System import readers
from smartcard.util import toHexString
from smartcard.Exceptions import NoCardException, ListReadersException
from smartcard.CardConnection import CardConnection
from smartcard.CardMonitoring import CardMonitor, CardObserver
from smartcard.scard import *
from colorama import init, Fore, Style
import platform

init()

def list_readers():
    try:
        pcsc_readers = readers()
        if not pcsc_readers:
            print("No smart card readers found.")
        else:
            print("Available readers:")
            for index, reader in enumerate(pcsc_readers):
                print(f"{index + 1}: {reader}")
        return pcsc_readers
    except ListReadersException as e:
        print(Fore.RED + f"Failed to list readers: {str(e)}" + Style.RESET_ALL)
        return []

def connect_to_reader(reader):
    # Standard connection (card present)
    connection = reader.createConnection()
    try:
        connection.connect(disposition=SCARD_LEAVE_CARD)
        print(Style.BRIGHT + Fore.GREEN + "Connected in standard mode" + Style.RESET_ALL)
        return connection, False
    except NoCardException:
        # Direct mode (no card)
        print(Style.BRIGHT + "No card detected, attempting direct mode..." + Style.RESET_ALL)
        try:
            if platform.system() != "Windows":
                # Handle direct connection within the PCSCCardConnection class
                connection.connect(protocol=CardConnection.RAW_protocol, mode=SCARD_SHARE_DIRECT)
                print(Style.BRIGHT + Fore.LIGHTMAGENTA_EX + "Connected using Direct Mode" + Style.RESET_ALL)
                return connection, True
            else:
                # Handle direct connection manually unfortunately
                direct_connection = reader.createConnection()
                hcontext = direct_connection.component.hcontext

                hresult, hcard, _ = SCardConnect(
                    hcontext,
                    reader.name,
                    SCARD_SHARE_DIRECT,
                    SCARD_PROTOCOL_UNSET
                )
                if hresult != SCARD_S_SUCCESS:
                    raise Exception(f"SCardConnect failed: {SCardGetErrorMessage(hresult)}")
                
                # Set hcard and disposition in PCSCCardConnection
                direct_connection.component.hcard = hcard
                direct_connection.component.hresult = hresult
                direct_connection.component.disposition = SCARD_LEAVE_CARD
                # direct_connection.connect(mode=SCARD_SHARE_DIRECT, protocol=SCARD_PROTOCOL_UNSET)
                print(Style.BRIGHT + Fore.LIGHTMAGENTA_EX + "Connected using Direct Mode" + Style.RESET_ALL)
                return direct_connection, True
        except Exception as e:
            print(Fore.RED + f"Direct Mode Failed: {str(e)}" + Style.RESET_ALL)
            return None, False
    except Exception as e:
        print(Fore.RED + f"Standard Connection Failed: {str(e)}" + Style.RESET_ALL)
        return None, False

def transmit_data(connection, data):
    apdu = [int(x, 16) for x in data.split()]
    response, sw1, sw2 = connection.transmit(apdu)
    print(Style.BRIGHT + Fore.YELLOW + f"Response: {toHexString(response)}" + Style.RESET_ALL)
    print(f"Status words: {sw1:02X} {sw2:02X}")

def control_data(connection, data):
    command_data = [int(x, 16) for x in data.split()]
    response_data = connection.control(SCARD_CTL_CODE(3500), command_data)
    print(Style.BRIGHT + Fore.YELLOW + f"Response: {toHexString(response_data)}" + Style.RESET_ALL)
    print(f"Status words: {response_data[-2]:02X} {response_data[-1]:02X}")

class CardObserverImpl(CardObserver):
    def update(self, observable, actions):
        addedcards, removedcards = actions
        for card in addedcards:
            print(Fore.LIGHTGREEN_EX + f"\n{card.reader}: Card present" + Style.RESET_ALL)
            print(toHexString(card.atr))
        for card in removedcards:
            print(Fore.LIGHTRED_EX + f"\n{card.reader}: Card absent" + Style.RESET_ALL)

if __name__ == "__main__":
    card_monitor = CardMonitor()
    observer = CardObserverImpl()
    card_monitor.addObserver(observer)

    while True:
        pcsc_readers = list_readers()
        if not pcsc_readers:
            print("No readers available, exiting...")
            break
        user_input = input(Style.BRIGHT + Fore.CYAN + "Select the reader to connect or press 'q' to exit: " + Style.RESET_ALL).strip()
        if user_input.lower() == 'q':
            break
        if not user_input:
            print("Please enter a number or 'q'")
            continue
        try:
            reader_index = int(user_input) - 1
            if 0 <= reader_index < len(pcsc_readers):
                connection, direct = connect_to_reader(pcsc_readers[reader_index])
                if connection:
                    try:
                        if direct:
                            data = input("Enter the control data to transmit (in hex, e.g., 'FF 00 01 00 00'): ")
                            control_data(connection, data)
                        else:
                            data = input("Enter the APDU to transmit (in hex, e.g., '80 84 00 00 08'): ")
                            transmit_data(connection, data)
                    except Exception as e:
                        print(Fore.RED + f"Transmit Failed: {str(e)}" + Style.RESET_ALL)
                    finally:
                        connection.disconnect()
                else:
                    print("Connection failed")
            else:
                print("Invalid reader index")
        except ValueError:
            print("Please enter a valid number or 'q'")

    card_monitor.deleteObserver(observer)